/*
 *   khexedit - Versatile hex editor
 *   Copyright (C) 1999  Espen Sand, espensa@online.no
 *
 *   The contents of this file is derived from kfontchooser.cpp of
 *   the KDE (1.x) libraries by Mark Donohoe and Stephan Kulow.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <qlayout.h>

#include <kapp.h>
#include <kcharsets.h>
#include <kglobal.h>
#include <klocale.h>

#include <X11/Xlib.h>

#include "dialogmargin.h"
#include "fontchooser.h"


CFontChooser::CFontChooser( QWidget *parent, const char *name )
  : QWidget( parent, name )
{
  QBoxLayout *topLayout = new QVBoxLayout( this, 3, INNER_SEP );

  mTypeFaceCombo = new QComboBox( false, this );
  mTypeFaceCombo->insertStringList( mFontList );
  connect( mTypeFaceCombo, SIGNAL(activated(const QString&)),
	   SLOT(slotSelectFont(const QString&)) );

  mTypeFaceLabel = new QLabel( mTypeFaceCombo, i18n("&Typeface"), this );
  mTypeFaceLabel->adjustSize();
  mTypeFaceLabel->setMinimumSize( mTypeFaceLabel->size() );
 
  topLayout->addWidget( mTypeFaceLabel );
  topLayout->addWidget( mTypeFaceCombo );

  mBoldCheck = new QCheckBox( i18n("&Bold"), this );
  mBoldCheck->setMinimumSize( mBoldCheck->sizeHint() );
  connect( mBoldCheck, SIGNAL(toggled(bool)), SLOT(slotFontBold(bool)) );
  topLayout->addWidget(mBoldCheck);
  
  mItalicCheck = new QCheckBox( i18n("&Italic"), this );
  mItalicCheck->setMinimumSize( mItalicCheck->sizeHint() );
  mItalicCheck->setChecked( mFont.italic() );
  connect( mItalicCheck, SIGNAL(toggled(bool)), SLOT(slotFontItalic(bool)) );  
  topLayout->addWidget( mItalicCheck );

  QGridLayout *gbox = new QGridLayout( 2, 3, INNER_SEP );
  if( gbox == 0 ) { return; }
  gbox->addColSpacing( 1, 10 );
  gbox->setColStretch( 2, 10 );
  topLayout->addLayout( gbox );
  
  mSizeSpin = new QSpinBox( this );
  mSizeSpin->setRange( 8, 16 );

  mSizeSpin->setMinimumWidth( fontMetrics().width('8')*10 );
  connect( mSizeSpin, SIGNAL(valueChanged(int)), SLOT( slotFontSize(int)) );

  mSizeLabel = new QLabel( mSizeSpin, i18n("&Size"), this );
  mSizeLabel->setMinimumSize( mSizeLabel->sizeHint() );

  mCharsetCombo = new QComboBox( false, this );
  mCharsetCombo->setMinimumWidth( mCharsetCombo->sizeHint().width() );
  mCharsetCombo->setInsertionPolicy( QComboBox::NoInsertion );
  connect( mCharsetCombo, SIGNAL(activated(const QString&)),
	   SLOT(slotCharset(const QString&)) );

  mCharSetLabel = new QLabel( mCharsetCombo, i18n("Cha&racter set"), this );
  mCharSetLabel->setMinimumWidth( mCharSetLabel->sizeHint().width() );

  gbox->addWidget( mSizeLabel, 0, 0 );
  gbox->addWidget( mSizeSpin, 1, 0 );
  gbox->addWidget( mCharSetLabel, 0, 2 );
  gbox->addWidget( mCharsetCombo, 1, 2 );
  topLayout->activate();


  getFontList( mFixedList, true );
  getFontList( mFontList, false );
  mFont = QFont( "helvetica", 12 );

  setFont( mFont, false );
  fillCharsetCombo();
}

void CFontChooser::setFont( QFont start_fnt, bool fixed )
{
  mFont = start_fnt;
  int i;
	
  mTypeFaceCombo->clear();
  if( fixed == true )
  {
    mTypeFaceCombo->insertStringList( mFixedList );
    QStringList::ConstIterator it;
    for( i=0, it = mFixedList.begin(); it != mFixedList.end(); i++, it++ )
    {
      if( *it == mFont.family() )
      {
	mTypeFaceCombo->setCurrentItem( i );
      }
    }
  }
  else
  {
    mTypeFaceCombo->insertStringList( mFontList );
    QStringList::ConstIterator it;
    for( i=0, it = mFontList.begin(); it != mFontList.end(); i++, it++ )
    {
      if( *it == mFont.family() )
      {
	mTypeFaceCombo->setCurrentItem( i );
      }
    }
  }
	
  mSizeSpin->setValue( mFont.pointSize() );
  mBoldCheck->setChecked( mFont.bold() );
  mItalicCheck->setChecked( mFont.italic() );
  fillCharsetCombo();
}

void CFontChooser::getFontList( QStringList &list, const char *pattern )
{
  int num;
  char **xFonts = XListFonts( qt_xdisplay(), pattern, 2000, &num );
  
  for( int i = 0; i < num; i++ )
  {
    addFont( list, xFonts[i] );
  }
  
  XFreeFontNames( xFonts );
}


void CFontChooser::getFontList( QStringList &list, bool fixed )
{
  //
  // Use KDE fonts if there is a KDE font list and check that the fonts
  // exist on the server where the desktop is running.
  //
	
  QStringList listSys, listKDE;
	
  if ( fixed == true ) 
  {
    getFontList( listSys, "-*-*-*-*-*-*-*-*-*-*-m-*-*-*" );
    //getFontList( listSys, "-*-*-*-*-*-*-*-*-*-*-c-*-*-*" );
  } 
  else
  {
    getFontList( listSys, "-*-*-*-*-*-*-*-*-*-*-p-*-*-*" );
  }
  
  if( kapp->getKDEFonts( listKDE ) == false ) 
  {
    list = listSys;
    return;
  }

  QStringList::ConstIterator it;
  for( it = listKDE.begin(); it != listKDE.end(); it++ )
  {
    QStringList::ConstIterator str = listSys.find( *it );
    if( str != 0 )
    {
      list.append( *str );
    }
  }
}


void CFontChooser::addFont( QStringList &list, const char *xfont )
{
  //
  // Move to the position after the two first '-' characters
  //
  const char *ptr = strchr( xfont, '-' );
  if( ptr != 0 )
  {
    ptr = strchr( ptr + 1, '-' );
  }
  if( ptr == 0 )
  {
    return;
  }

  QString font = ptr + 1;
  int pos = font.find( '-' );
  if( pos == -1 )
  {
    return;
  }

  font.truncate( pos );
  if ( font.find( "open look", 0, false ) >= 0 )
  {
    return;
  }

  QStringList::ConstIterator it;
  for( it = list.begin(); it != list.end(); it++ )
  {
    if( *it == font )
    {
      return;
    }
  }

  list.append( font );
  list.sort();
}


void CFontChooser::fillCharsetCombo( void )
{
  mCharsetCombo->clear();
  KCharsets *charsets = KGlobal::charsets(); 
  QStringList list = charsets->availableCharsetNames( mFont.family() );
  
  mCharsetCombo->insertItem( i18n("default") );
  mCharsetCombo->insertStringList( list );
  QString charset = charsets->name( mFont );
  for( int i=0; i < mCharsetCombo->count(); i++ )
  {
    if( charset == mCharsetCombo->text(i) )
    {
      mCharsetCombo->setCurrentItem(i);
      break;
    }
  }
}


void CFontChooser::slotCharset(const QString &name )
{
  KCharsets *charsets = KGlobal::charsets(); 
  if( name == i18n("default") )
  {
    charsets->setQFont( mFont, KGlobal::locale()->charset() );
  }
  else
  {
    charsets->setQFont( mFont, name );
  }
  emit fontChanged( mFont );
}


void CFontChooser::slotSelectFont( const QString &fontName )
{
  mFont.setFamily( fontName );	
  fillCharsetCombo();	
  emit fontChanged( mFont );
}


void CFontChooser::slotFontSize( int s )
{
  mFont.setPointSize( s );
  emit fontChanged( mFont );
}


void CFontChooser::slotFontBold( bool b )
{
  mFont.setBold( b );
  emit fontChanged( mFont );
}


void CFontChooser::slotFontItalic( bool i )
{
  mFont.setItalic( i );
  emit fontChanged( mFont );
}


void CFontChooser::setEnabled( bool state )
{
  mTypeFaceLabel->setEnabled( state );
  mSizeLabel->setEnabled( state );
  mCharSetLabel->setEnabled( state );

  mSizeSpin->setEnabled( state );
  mTypeFaceCombo->setEnabled( state );
  mCharsetCombo->setEnabled( state );
  mBoldCheck->setEnabled( state );
  mItalicCheck->setEnabled( state );
}


const QFont &CFontChooser::font( void )
{
  return( mFont );
}
