/*
 *   khexedit - Versatile hex editor
 *   Copyright (C) 1999  Espen Sand, espensa@online.no
 *   This file is based on the work by F. Zigterman, fzr@dds.nl
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef _TOPLEVEL_H_
#define _TOPLEVEL_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <kapp.h>
#include <kstdaccel.h>
#include <ktmainwindow.h>

#include "aboutdialog.h"
#include "dragsource.h"
#include "hexbuffer.h"
#include "hexmanagerwidget.h"
#include "statusbarprogress.h"

class KHexEdit : public KTMainWindow
{
  enum EMenuConstant
  {
    permission_readonly = 1,
    permission_allowResize,
    view_hexadecimal,
    view_octal,
    view_binary,
    view_text,
    view_offset_area,
    view_text_area,
    view_offset_as_decimal,
    view_case_data,
    view_case_offset,
    view_lock_width,
    encoding_default,
    encoding_us_ascii,
    encoding_ebcdic,
    conversion_hide,
    conversion_window,
    conversion_editor,
    option_edit,
    option_favorites,
    option_toolbar,
    option_conversion,
    option_statusbar,
    help_help,
    help_aboutKhexedit,
    help_aboutKDE,
    toolbar_Copy,
    toolbar_Cut,
    toolbar_Undo,
    toolbar_Redo,
    toolbar_rw,
    toolbar_cancelOperation,
  };

  Q_OBJECT

  public:
    KHexEdit( void );
    ~KHexEdit( void );

    inline bool isValid( void );
    inline void addStartupFile( const QString &fileName );
    inline void setStartupOffset( uint offset );

  public slots:
    void recentFile( int id );
    void newWindow( void );
    void pasteNewWindow( void );
    void closeWindow( void );
    void closeProgram( void );
    void statusBarPressed( int id );

    void operationChanged( bool state );
    void cursorChanged( SCursorState &state );
    void fileModified( SFileState &state );
    void layoutChanged( const SDisplayLayout &layout );
    void inputModeChanged( const SDisplayInputMode &mode );
    void bookmarkChanged( QList<SCursorOffset> &list );
    void removeRecentFile( const QString &fileName );
    void renameRecentFile( QString &curName, QString &newName );

    void fileActive( QString &url );
    void fileRename( QString &curName, QString &newName );
    void fileClosed( QString &url );
    void readConfiguration( void );
    void writeConfiguration( void );
    void editMode( CHexBuffer::EEditMode editMode );
    void encodingChanged( const SEncodeState &state );
    void textWidth( uint width );
  
    void viewMenuCB( int id );
    void encodingMenuCB( int id );
    void bookmarkMenuCB( int id );
    void bookmarkPopupCB( int id );
    void documentMenuCB( int id );
    void optionMenuCB( int id );
   
    void invokeHelp( void );
    void showAboutDialog( void );

  protected:
    virtual bool queryExit( void );
    virtual bool queryClose( void );
    void timerEvent( QTimerEvent *e );

  private slots:
    void enableMenuIcon( bool state );
    void removeRecentFiles( void );
    void conversionClosed( void );
    

  private:
    void open( QStringList &fileList, QStringList &offsetList );
    void initialize( bool openFiles );
    void setConversionVisibility( 
      CHexManagerWidget::EConversionPosition position );
    void addRecentFile( QString &fileName );
    bool closeAllWindow( void );
    void setUndoState( uint undoState );
    void setSelectionState( bool selectionValid );
    void addDocument( QString &fileName );
    void removeDocument( QString &fileName );
    void renameDocument( QString &curName, QString &newName );

    void makeExitFileList( void );
    void writeConfiguration( KConfig &config );
    void readConfiguration( KConfig &config );
    void setStartupGeometry( void );
    void updateCaption( const QString &url );
    bool eventFilter( QObject *obj, QEvent *event );

    inline CHexEditorWidget *editor( void );
    inline CHexToolWidget *util( void );
    inline CHexViewWidget *textView( void );
    inline int acceleratorNumKey( uint index );
 
  private:
    static QList<KHexEdit> mWindowList;
    QList<QString> mDocumentList;
    CHexManagerWidget *mManager;
 
    QStringList mStartupFileList;   // Files to automatically open on startup
    QStringList mStartupOffsetList; // Start offset for those files.

    QStringList mExitFileList;      // Stores "Recent" file order on exit.
    QStringList mExitOffsetList;    // Stores "Recent" file offset on exit. 

    QStringList mRecentOffsetList;  // Values read from config. file
    uint mStartupOffset;            // Value read from commandline

    QPopupMenu *mFilePermission;
    QPopupMenu *mFile;
    QPopupMenu *mRecent;
    QPopupMenu *mEdit;
    QPopupMenu *mEditSpecial;
    QPopupMenu *mView;
    QPopupMenu *mEncoding;
    QPopupMenu *mBookmark;
    QPopupMenu *mDocuments;
    QPopupMenu *mTools;
    QPopupMenu *mConversion;
    QPopupMenu *mOptions;
    QPopupMenu *mHelp;
    QPopupMenu *mEditorPopup;
    QPopupMenu *mBookmarkPopup;

    int mFileNewDocumentMB, mFileOpenMB, mFileReloadMB, mFileSaveMB;
    int mFilePrintMB, mFileCancelOperationMB, mFileQuitMB;
    int mPermissionReadOnlyMB, mPermissionResizeMB;
    int mEditUndoMB, mEditRedoMB, mEditCopyMB, mEditPasteMB, mEditCutMB;
    int mEditSpecialCopyText1MB, mEditSpecialCopyText2MB;
    int mEditFindMB;
    int mViewHexMB, mViewOctalMB, mViewBinaryMB, mViewTextOnlyMB;
    int mViewOffsetColumnMB, mViewTextAreaMB, mViewOffsetAsDecimalMB;
    int mViewBigCaseDataMB, mViewBigCaseOffsetMB;
    int mEncodingDefaultMB, mEncodingUsAsciiMB, mEncodingEbcdicMB;
    int mOptionToolBarMB, mOptionConversionMB, mOptionStatusBarMB;
    int mConversionHideMB, mConversionWindowMB, mConversionEditorMB;
    int mPopHexMB, mPopOctalMB, mPopBinaryMB, mPopTextOnlyMB;
    int mPopReadOnlyMB, mPopResizeMB;
    
    CUrlLabel *mUrlLabel;
    CAboutDialog *mAboutDialog;

    bool mValid;
    bool mShowToolBar;
    bool mShowStatusBar;
    bool mSelectionValid;
    uint mUndoState;
};


inline bool KHexEdit::isValid( void )
{
  return( mValid );
}

inline void KHexEdit::addStartupFile( const QString &fileName )
{
  mStartupFileList.prepend( fileName );
  mStartupOffsetList.prepend( QString("%1").arg(mStartupOffset,0,16) );
  mStartupOffset = 0;
}

inline void KHexEdit::setStartupOffset( uint offset )
{
  mStartupOffset = offset;
}

inline CHexEditorWidget *KHexEdit::editor( void )
{
  return( mManager->editor() );
}

inline CHexViewWidget *KHexEdit::textView( void )
{
  return( mManager->editor()->view() );
}

inline CHexToolWidget *KHexEdit::util( void )
{
  return( mManager->util() );
}

int KHexEdit::acceleratorNumKey( uint index )
{
  static int keys[9] = 
  {
    Key_1, Key_2, Key_3, Key_4, Key_5, Key_6, Key_7, Key_8, Key_9
  };
  return( index >= 9 ? -1 : keys[index] );
}



#endif


